import os
import random
import zipfile
from datetime import datetime, timezone, timedelta
from pathlib import Path

import pytest
import lxml.etree as ET

SIMPLIFIED_ANNOTATION = b"""<?xml version="1.0" encoding="UTF-8"?>
<product>
  <swathTiming>
    <linesPerBurst>1000</linesPerBurst>
    <samplesPerBurst>20000</samplesPerBurst>
    <burstList count="2">
      <burst>
        <azimuthTime>2022-01-01T00:00:00.000000</azimuthTime>
        <azimuthAnxTime>1.00e+03</azimuthAnxTime>
        <sensingTime>2022-00-00T00:00:00.000000</sensingTime>
        <byteOffset>10</byteOffset>
        <firstValidSample count="9">-1 -1 10 10 10 10 10 -1 -1</firstValidSample>
        <lastValidSample count="9">-1 -1 100 100 100 100 100 -1 -1</lastValidSample>
      </burst>
      <burst>
        <azimuthTime>2022-01-01T00:00:01.000000</azimuthTime>
        <azimuthAnxTime>2.00+03</azimuthAnxTime>
        <sensingTime>2022-01-01T00:00:01.000000</sensingTime>
        <byteOffset>110</byteOffset>
        <firstValidSample count="9">-1 -1 10 10 10 10 10 -1 -1</firstValidSample>
        <lastValidSample count="9">-1 -1 100 100 100 100 100 -1 -1</lastValidSample>
      </burst>
    </burstList>
  </swathTiming>
  <geolocationGrid>
    <geolocationGridPointList count="2">
      <geolocationGridPoint>
        <azimuthTime>2022-01-01T00:00:00.000000</azimuthTime>
        <slantRangeTime>5.00e-03</slantRangeTime>
        <line>0</line>
        <pixel>0</pixel>
        <latitude>2.00e+01</latitude>
        <longitude>5.00e+01</longitude>
        <height>1.00e+03</height>
      </geolocationGridPoint>
      <geolocationGridPoint>
        <azimuthTime>2020-06-04T02:22:53.625336</azimuthTime>
        <slantRangeTime>5.5e-03</slantRangeTime>
        <line>0</line>
        <pixel>1273</pixel>
        <latitude>2.50e+01</latitude>
        <longitude>5.50e+01</longitude>
        <height>1.5e+03</height>
      </geolocationGridPoint>
    </geolocationGridPointList>
  </geolocationGrid>
</product>
"""


@pytest.fixture()
def annotation_xml():
    return ET.fromstring(SIMPLIFIED_ANNOTATION)


@pytest.fixture()
def example_credentials():
    expire_time = datetime.now(timezone.utc) + timedelta(hours=1)
    example_credentials = {
        'accessKeyId': 'foo',
        'secretAccessKey': 'bar',
        'sessionToken': 'baz',
        'expiration': expire_time.isoformat(),
    }
    yield example_credentials


@pytest.fixture()
def zip_stored(tmpdir):
    file_path = Path(tmpdir) / 'test_file.txt'
    file_path.write_text('Hello World!')
    zip_path = Path(tmpdir) / 'test_file.zip'
    with zipfile.ZipFile(zip_path, 'w', compression=zipfile.ZIP_STORED) as zip_file:
        zip_file.write(file_path)
    yield zip_path, file_path.name
    zip_path.unlink()
    file_path.unlink()


@pytest.fixture(scope='module')
def data():
    # Can't use os.random directly because there needs to be some
    # repitition in order for compression to be effective
    words = [os.urandom(8) for _ in range(1000)]
    out = b''.join([random.choice(words) for _ in range(524288)])
    return out


@pytest.fixture()
def zip_deflated(tmpdir, data):
    file_path = Path(tmpdir) / 'foo-iw2-bar-vv-baz.tiff'
    file_path.write_bytes(data)
    zip_path = Path(tmpdir) / 'SLCNAME.zip'
    with zipfile.ZipFile(zip_path, 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=7) as zip_file:
        zip_file.write(file_path)
    yield zip_path, file_path.name
    zip_path.unlink()
    file_path.unlink()



import json
import math
import os
import zipfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterable

import boto3
import botocore
import requests
import zran
from tqdm import tqdm


KB = 1024
MB = 1024 * KB
SENTINEL_DISTRIBUTION_URL = 'https://sentinel1.asf.alaska.edu'
BUCKET = 'asf-ngap2w-p-s1-slc-7b420b89'


def check_if_in_aws_and_region(region='us-west-2'):
    using_ec2 = False
    try:
        with open('/var/lib/cloud/instance/datasource') as f:
            line = f.readlines()
            if 'DataSourceEc2' in line[0]:
                using_ec2 = True
    except FileNotFoundError:
        pass

    using_lambda = False
    if 'AWS_LAMBDA_FUNCTION_NAME' in os.environ:
        using_lambda = True

    in_aws = using_ec2 or using_lambda

    if not in_aws:
        return False

    if not boto3.Session().region_name == region:
        return False

    return True


@dataclass(frozen=True)
class GeoControlPoint:
    pixel: int
    line: int
    lon: float
    lat: float
    hgt: float


# TODO can also be np.array
@dataclass(frozen=True)
class Offset:
    start: int
    stop: int


@dataclass(frozen=True)
class Window:
    xstart: int
    xend: int
    ystart: int
    yend: int


@dataclass(frozen=True)
class BurstMetadata:
    name: str
    slc: str
    swath: str
    burst_index: int
    shape: Iterable[int]  # n_row, n_column
    index_offset: Offset
    uncompressed_offset: Offset
    annotation_offset: Offset
    manifest_offset: Offset

    def to_dict(self):
        dictionary = {
            'name': self.name,
            'slc': self.slc,
            'swath': self.swath,
            'burst_index': self.burst_index,
            'n_rows': int(self.shape[0]),
            'n_columns': int(self.shape[1]),
            'index_offset': {'start': int(self.index_offset.start), 'stop': int(self.index_offset.stop)},
            'uncompressed_offset': {
                'start': int(self.uncompressed_offset.start),
                'stop': int(self.uncompressed_offset.stop),
            },
            'annotation_offset': {
                'start': int(self.annotation_offset.start),
                'stop': int(self.annotation_offset.stop),
            },
            'manifest_offset': {
                'start': int(self.manifest_offset.start),
                'stop': int(self.manifest_offset.stop),
            },
        }
        return dictionary


@dataclass(frozen=True)
class XmlMetadata:
    name: str
    slc: str
    offset: Offset

    def to_dict(self):
        return {self.slc: {self.name: {'offset_start': self.offset.start, 'offset_stop': self.offset.stop}}}


def calculate_range_parameters(total_size: int, offset: int, chunk_size: int) -> Iterable[str]:
    """Calculate range parameters for HTTP range requests.
    Useful when downloading large files in chunks.

    Args:
        total_size: total size of request
        offset: offset to start request
        chunk_size: size of each chunk

    Returns:
        list of range parameters
    """
    num_parts = int(math.ceil(total_size / float(chunk_size)))
    range_params = []
    for part_index in range(num_parts):
        start_range = (part_index * chunk_size) + offset
        if part_index == num_parts - 1:
            end_range = str(total_size + offset - 1)
        else:
            end_range = start_range + chunk_size - 1

        range_params.append(f'bytes={start_range}-{end_range}')
    return range_params


def get_tmp_access_keys(save_path: Path = Path('./credentials.json'), edl_token: str = None) -> dict:
    """Get temporary AWS access keys for direct
    access to ASF data in S3.

    Args:
        save_path: path to save credentials to

    Returns:
        dictionary of credentials
    """
    if not edl_token:
        edl_token = os.environ['EDL_TOKEN']
    resp = requests.get(
        'https://sentinel1.asf.alaska.edu/s3credentials',
        headers={'Authorization': f'Bearer {edl_token}'},
    )
    resp.raise_for_status()
    save_path.write_bytes(resp.content)
    return resp.json()


def get_credentials(edl_token: str = None, working_dir=Path('.')) -> dict:
    """Gets temporary ASF AWS credentials from
    file or request new credentials if credentials
    are not present or expired.

    Returns:
        dictionary of credentials
    """
    credential_file = working_dir / 'credentials.json'
    if not credential_file.exists():
        credentials = get_tmp_access_keys(credential_file, edl_token)
        return credentials

    credentials = json.loads(credential_file.read_text())
    expiration_time = datetime.fromisoformat(credentials['expiration'])
    current_time = datetime.now(timezone.utc)

    if current_time >= expiration_time:
        credentials = get_tmp_access_keys(credential_file, edl_token)

    return credentials


def lambda_get_credentials(edl_token, working_dir, client, bucket, key):
    """Gets temporary ASF AWS credentials for a lambda function.
    Checks if credentials exist in S3 and are not expired. If neither are true,
    a new version of the credentials is uploaded to S3.

    Args:
        edl_token: EDL token
        working_dir: working directory
        client: boto3 client
        bucket: S3 bucket
        key: S3 key
    """
    credential_file = working_dir / 'credentials.json'
    credentials_need_update = False

    try:
        client.download_file(bucket, key, credential_file)
    except botocore.exceptions.ClientError:
        print('The credentials do not exist, will create.')
        credentials_need_update = True
        get_tmp_access_keys(credential_file, edl_token)

    credentials = json.loads(credential_file.read_text())
    expiration_time = datetime.fromisoformat(credentials['expiration'])
    current_time = datetime.now(timezone.utc)

    if current_time >= expiration_time:
        credentials_need_update = True
        get_tmp_access_keys(credential_file, edl_token)

    if credentials_need_update:
        client.upload_file(credential_file, bucket, key)


def get_download_url(scene: str) -> str:
    """Get download url for Sentinel-1 scene.

    Args:
        scene: scene name

    Returns:
        Scene url
    """
    mission = scene[0] + scene[2]
    product_type = scene[7:10]
    if product_type == 'GRD':
        product_type += '_' + scene[10] + scene[14]
    url = f'{SENTINEL_DISTRIBUTION_URL}/{product_type}/{mission}/{scene}.zip'
    return url


def s3_range_get(client: boto3.client, url: str, range_header: str) -> bytes:
    """Get a range of bytes from an S1 SLC file in ASF's archive.
    Used in threading to download a large file in chunks.

    Args:
        client: boto3 S3 client
        url: url location of SLC file
        range_header: range header string

    Returns:
        bytes of object
    """
    key = Path(url).name
    resp = client.get_object(Bucket=BUCKET, Key=key, Range=range_header)
    body = resp['Body'].read()
    return body


def http_range_get(client: requests.sessions.Session, url: str, range_header: str) -> bytes:
    """Get a range of bytes from an S1 SLC file in ASF's archive.
    Used in threading to download a large file in chunks.

    Args:
        client: requests session
        url: url location of SLC file
        range_header: range header string

    Returns:
        bytes of object
    """
    resp = client.get(url, headers={'Range': range_header})
    body = resp.content
    return body


def setup_download_client(strategy: str = 's3', edl_token: str = None, working_dir: Path = Path('.')) -> bytes:
    """Create client and range_get_func for downloading from SLC archive based on strategy (s3 | http).

    Args:
        strategy: strategy to use for download (s3 | http) s3 only works if runnning from us-west-2 region
        edl_token: EDL token for downloading from ASF's archive, if None will assume token is specified
                   in environment variable EDL_TOKEN
        working_dir: working directory where temparary credentials will be stored

    Returns:
        S3 client or http client, and matching *_range_get function
    """
    if strategy == 's3':
        creds = get_credentials(edl_token, working_dir)
        client = boto3.client(
            's3',
            aws_access_key_id=creds['accessKeyId'],
            aws_secret_access_key=creds['secretAccessKey'],
            aws_session_token=creds['sessionToken'],
        )
        range_get = s3_range_get

    elif strategy == 'http':
        client = requests.session()
        range_get = http_range_get

    return client, range_get


def download_slc(scene: str, edl_token: str = None, working_dir=Path('.'), strategy='s3') -> str:
    """Download an SLC zip file from ASF.

    Args:
        scene: SLC name (no exstension)
        strategy: strategy to use for download (s3 | http)
            s3 only works if runnning from us-west-2 region
    """
    zip_name = f'{scene}.zip'
    url = get_download_url(scene)

    client, _ = setup_download_client(strategy, edl_token, working_dir)
    if strategy == 's3':
        metadata = client.head_object(Bucket=BUCKET, Key=zip_name)
        total_length = int(metadata.get('ContentLength', 0))
        with tqdm(total=total_length, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
            with open(working_dir / zip_name, 'wb') as f:
                client.download_fileobj(BUCKET, zip_name, f, Callback=pbar.update)

    elif strategy == 'http':
        with client.get(url, stream=True) as s:
            s.raise_for_status()
            total = int(s.headers.get('content-length', 0))
            with open(working_dir / zip_name, 'wb') as f:
                with tqdm(unit='B', unit_scale=True, unit_divisor=1024, total=total) as pbar:
                    for chunk in s.iter_content(chunk_size=10 * MB):
                        if chunk:
                            f.write(chunk)
                            pbar.update(len(chunk))
        client.close()


def get_zip_compressed_offset(zip_path: str, zinfo: zipfile.ZipInfo) -> Offset:
    """Get the byte offset (beginning byte and end byte inclusive) for a member
    of a zip archive.

    Args:
        zip_path: path to zip file on disk
        zinfo: the ZipInfo object associated with the zip member to get offset for

    Returns:
        byte offset (start and end) for zip member
    """
    with open(zip_path, 'rb') as f:
        f.seek(zinfo.header_offset)
        data = f.read(30)

    n = int.from_bytes(data[26:28], 'little')
    m = int.from_bytes(data[28:30], 'little')

    data_start = zinfo.header_offset + n + m + 30
    data_stop = data_start + zinfo.compress_size
    return Offset(data_start, data_stop)


class ZipIndexer:
    """Class for creating dflidx indexes for zip archive members
    that are compatible with the zran library.
    """

    def __init__(self, zip_path: str, member_name: str, spacing: int = 2**20):
        self.zip_path = zip_path
        self.spacing = spacing
        self.member_name = member_name
        self.index = None

    def create_full_dflidx(self):
        with zipfile.ZipFile(self.zip_path) as f:
            zinfo = [x for x in f.infolist() if self.member_name in x.filename][0]

        self.file_offset = get_zip_compressed_offset(self.zip_path, zinfo)
        with open(self.zip_path, 'rb') as f:
            f.seek(self.file_offset.start)
            self.body = f.read(self.file_offset.stop - self.file_offset.start)

        self.index = zran.Index.create_index(self.body, span=self.spacing)

    def subset_dflidx(self, locations: Iterable[int], end_location: int) -> Iterable:
        """Build base DFLIDX index for a Zip member file that has
        been compresed using zlib (DEFLATE).

        Args:
            locations: A list of uncompressed locations to be included in the new index.
                       The closes point before each location will be selected.
            end_location: The uncompressed endpoint of the index. Used to determine file size.

        Returns:
            bytes of dflidx, and compressed range of member in zip archive
        """
        compressed_offset, uncompressed_offset, modified_index = self.index.create_modified_index(
            locations, end_location
        )
        compressed_offset = Offset(
            compressed_offset[0] + self.file_offset.start, compressed_offset[1] + self.file_offset.start
        )
        uncompressed_offset = Offset(uncompressed_offset[0], uncompressed_offset[1])
        return compressed_offset, uncompressed_offset, modified_index



import json
import os
import boto3

lamb = boto3.client('lambda')

edl_token = os.environ['EDL_TOKEN']
bursts = [
    'S1B_IW_SLC__1SDV_20201115T162313_20201115T162340_024278_02E29D_5C54_IW3_VV_4',
]

function_arn = (
    'arn:aws:lambda:us-west-2:050846374571:function:ffwilliams2-test-indexer-BurstExtractFunction-kcaujwt92zHk'
)

results = []
for burst in bursts:
    payload = {'burst': burst, 'edl_token': edl_token}
    result = lamb.invoke(FunctionName=function_arn, InvocationType='Event', Payload=json.dumps(payload))
    if result['StatusCode'] != 202:
        print(f'Invocation failed with for {burst} with status code {result["StatusCode"]}')
    else:
        print(f'{burst} successfully submitted!')



import json
import zipfile
from datetime import datetime, timezone, timedelta
from pathlib import Path
from unittest.mock import patch

import botocore
import requests
import zran

from index_safe import utils


def test_calculate_range_parameters():
    params = utils.calculate_range_parameters(100, 50, 10)
    assert int(params[-1][-3:]) == 149
    assert int(params[0][6:8]) == 50
    assert len(params) == 10

    spacing = int(params[1][-2:]) - int(params[1][6:8])
    spacing += 1  # range reqests are inclusive
    assert spacing == 10


def test_get_credentials(tmpdir, example_credentials):
    credential_file = Path(tmpdir) / 'credentials.json'
    credential_file.write_text(json.dumps(example_credentials))
    output_file = utils.get_credentials(working_dir=Path(tmpdir))
    assert output_file['accessKeyId'] == 'foo'

    example_credentials['expiration'] = (datetime.now(timezone.utc) - timedelta(hours=1)).isoformat()
    credential_file.write_text(json.dumps(example_credentials))
    with patch('index_safe.utils.get_tmp_access_keys', return_value={}) as mock_func:
        utils.get_credentials(working_dir=Path(tmpdir))
        mock_func.assert_called_once()

    credential_file.unlink()
    with patch('index_safe.utils.get_tmp_access_keys', return_value={}) as mock_func:
        utils.get_credentials(working_dir=Path(tmpdir))
        mock_func.assert_called_once()


def test_get_download_url():
    safe_name = 'S1A_IW_SLC__1SDV_20230812T001358_20230812T001424_049835_05FE51_CAB0'
    real_url = f'https://sentinel1.asf.alaska.edu/SLC/SA/{safe_name}.zip'
    created_url = utils.get_download_url(safe_name)
    assert real_url == created_url


def test_setup_download_client(example_credentials):
    client, range_get = utils.setup_download_client(strategy='http')
    assert isinstance(client, requests.sessions.Session)
    assert range_get is utils.http_range_get

    with patch('index_safe.utils.get_tmp_access_keys', return_value=example_credentials):
        client, range_get = utils.setup_download_client(strategy='s3')
        assert isinstance(client, botocore.client.BaseClient)
        assert range_get is utils.s3_range_get


def test_get_zip_compressed_offset(zip_stored):
    zip_path, member_name = zip_stored
    with zipfile.ZipFile(zip_path) as f:
        zinfo = [x for x in f.infolist() if member_name in x.filename][0]

    offset = utils.get_zip_compressed_offset(zip_path, zinfo)
    with open(zip_path, 'rb') as f:
        f.seek(offset.start)
        output = f.read(offset.stop - offset.start)

    assert b'Hello World!' == output


def test_get_zip_indexer():
    index = utils.ZipIndexer('test.zip', 'test.txt')
    assert index.zip_path == 'test.zip'
    assert index.member_name == 'test.txt'
    assert index.spacing == 2**20
    assert index.index is None


def test_create_full_dflidx(zip_deflated):
    zip_path, member_name = zip_deflated
    indexer = utils.ZipIndexer(zip_path, member_name)
    indexer.create_full_dflidx()
    assert isinstance(indexer.index, zran.Index)
    assert indexer.index.have > 0
    first_point = indexer.index.points[0]
    assert first_point.outloc == 0
    assert first_point.inloc == 0
    assert first_point.bits == 0


def test_subset_dflidx(zip_deflated, data):
    zip_path, member_name = zip_deflated
    indexer = utils.ZipIndexer(zip_path, member_name, spacing=2**18)
    indexer.create_full_dflidx()
    step = 2**18
    compressed_offset, uncompressed_offset, modified_index = indexer.subset_dflidx([2 * step, 4 * step], 6 * step)
    assert modified_index.have == 2

    first_point = modified_index.points[0]
    assert first_point.outloc == 0
    assert first_point.inloc == 1  # to account for non-zero point.bits

    with open(zip_path, 'rb') as f:
        f.seek(compressed_offset.start)
        compressed_test_data = f.read(compressed_offset.stop - compressed_offset.start)
    test_data = zran.decompress(compressed_test_data, modified_index, 0, modified_index.uncompressed_size)
    assert test_data == data[uncompressed_offset.start : uncompressed_offset.stop]



import base64
import io
import json
import os
import tempfile
import lxml.etree as ET
import zipfile
from argparse import ArgumentParser
from pathlib import Path
from typing import Iterable, Union

import boto3
from tqdm import tqdm


try:
    from index_safe import utils
except ModuleNotFoundError:
    import utils


def load_annotation_data(zipped_safe_path: str, tiff_name: str) -> ET.Element:
    """Load annotation XML from zipped SAFE.

    Args:
        zipped_safe_path: path to a SAFE zip containing the annotation XML
        tiff_name: name of the tiff to extract annotation XML for

    Returns:
        xml: ElementTree object containing annotation XML
    """
    safe_name = Path(zipped_safe_path).with_suffix('.SAFE').name
    annotation_path = Path(safe_name) / 'annotation' / Path(tiff_name).with_suffix('.xml')
    with zipfile.ZipFile(zipped_safe_path) as f:
        annotation_bytes = f.read(str(annotation_path))
    xml = ET.parse(io.BytesIO(annotation_bytes)).getroot()
    return xml


def get_burst_annotation_data(xml: ET.Element) -> Iterable:
    """Obtain the information needed to extract a burst that is contained within
    a SAFE annotation XML.

    Args:
        xml: ElementTree object containing annotation XML

    Returns:
        burst_shape: numpy-style tuple of burst array size (n rows, n columns)
        burst_offsets: uncompressed byte offsets for the bursts contained within
            a swath
    """
    burst_xmls = xml.findall('.//{*}burst')
    n_lines = int(xml.findtext('.//{*}linesPerBurst'))
    n_samples = int(xml.findtext('.//{*}samplesPerBurst'))
    burst_shape = (n_lines, n_samples)  #  y, x for numpy
    burst_starts = [int(x.findtext('.//{*}byteOffset')) for x in burst_xmls]
    burst_lengths = burst_starts[1] - burst_starts[0]
    burst_offsets = [utils.Offset(x, x + burst_lengths) for x in burst_starts]
    return burst_shape, burst_offsets


def create_xml_metadata(zipped_safe_path: str, zinfo: zipfile.ZipInfo) -> utils.XmlMetadata:
    """Create object containing information needed to download metadata XML file from
    compressed file directly.

    Args:
        zipped_safe_path: Path to zipped SAFE
        zinfo: ZipInfo object for desired XML

    Returns:
        XmlMetadata object containing offsets needed to download XML
    """
    slc_name = Path(zipped_safe_path).with_suffix('').name
    name = Path(zinfo.filename).name
    offset = utils.get_zip_compressed_offset(zipped_safe_path, zinfo)
    return utils.XmlMetadata(name, slc_name, offset)


def create_burst_name(slc_name: str, swath_name: str, burst_index: str) -> str:
    """Create name for a burst tiff.

    Args:
        slc_name: Name of SLC
        swath_name: Name of swath
        burst_index: Zero-indexed burst number in swath

    Returns:
        Name of burst
    """
    _, swath, _, polarization, *_ = swath_name.split('-')
    all_parts = [slc_name, swath.upper(), polarization.upper(), str(burst_index)]
    return '_'.join(all_parts) + '.tiff'


def create_burst_dflidx(
    indexer: utils.ZipIndexer, burst_offset: utils.Offset
) -> Iterable[Union[utils.Offset, utils.Offset, bytes]]:
    """Create a burst-specific zran index containing information needed to download burst tiff from compressed

    Args:
        indexer: ZipIndexer object for the SAFE
        burst_offset: Uncompressed offset of burst within SAFE

    Returns:
        compressed_offset: Compressed offset of zran window for burst within SAFE
        index_burst_offset: Uncompressed offset of burst within zran index
        dflidx: zran index file for burst
    """
    compressed_offset, uncompressed_offset, modified_index = indexer.subset_dflidx(
        locations=[burst_offset.start], end_location=burst_offset.stop
    )
    index_burst_offset = utils.Offset(
        burst_offset.start - uncompressed_offset.start, burst_offset.stop - uncompressed_offset.start
    )
    dflidx = modified_index.create_index_file()
    return compressed_offset, index_burst_offset, dflidx


def create_index(
    zipped_safe_path: str,
    tiff_name: str,
    annotation_offset: utils.Offset,
    manifest_offset: utils.Offset,
    output_json: bool = True,
) -> Union[dict, bytes]:
    """Create a burst-specific index containing information needed to download burst tiff from compressed
    SAFE file directly, and remove invalid data.

    Args:
        zipped_safe_path: Path to zipped SAFE
        tiff_name: Name of the swath tiff
        annotation_offset: Uncompressed offset of annotation XML within SAFE
        manifest_offset: Uncompressed offset of manifest XML within SAFE
        output_json: Whether to output index as json or bytes

    Returns:
        dictionary or bytes object containing information needed to download and remove invalid data
    """
    slc_name = Path(zipped_safe_path).with_suffix('').name
    swath_name = tiff_name.split('-')[1].upper()

    xml = load_annotation_data(zipped_safe_path, tiff_name)
    burst_shape, burst_offsets = get_burst_annotation_data(xml)
    swath_level_info = {
        'slc': slc_name,
        'swath': swath_name,
        'shape': burst_shape,
        'annotation_offset': annotation_offset,
        'manifest_offset': manifest_offset,
    }

    bursts = {}
    indexer = utils.ZipIndexer(zipped_safe_path, tiff_name)
    indexer.create_full_dflidx()
    for burst_index, burst_offset in enumerate(burst_offsets):
        burst_name = create_burst_name(slc_name, tiff_name, burst_index)
        compressed_offset, index_burst_offset, dflidx = create_burst_dflidx(indexer, burst_offset)
        burst = utils.BurstMetadata(
            name=burst_name,
            burst_index=burst_index,
            index_offset=compressed_offset,
            uncompressed_offset=index_burst_offset,
            **swath_level_info,
        )

        if output_json:
            bstidx_name = Path(burst_name).with_suffix('.json').name
            burst_dictionary = burst.to_dict()
            burst_dictionary['dflidx_b64'] = base64.b64encode(dflidx).decode('utf-8')
            bursts[bstidx_name] = burst_dictionary
        else:
            bstidx_name = Path(burst_name).with_suffix('.bstidx').name
            bursts[bstidx_name] = burst.to_bytes() + dflidx
    return bursts


def save_xml_metadata_as_json(entries: Iterable[utils.XmlMetadata], out_name: str) -> str:
    """Save a list of XmlMetadata objects as a json.

    Args:
        entries: List of metadata objects to be included
        out_name: Path/name to save json at

    Returns:
        Path/name where json was saved
    """
    slc = entries[0].slc
    metadata_dicts = [entry.to_dict()[slc] for entry in entries]
    combined_dict = {}
    for metadata_dict in metadata_dicts:
        combined_dict.update(metadata_dict)

    with open(out_name, 'w') as json_file:
        json.dump({slc: combined_dict}, json_file)
    return out_name


def save_burst_metadata_as_json(burst_metadata_dict: dict, working_dir: Path) -> Path:
    """Save a dictionary of burst metadata objects as jsons.

    Args:
        burst_metadata_dict: Dictionary of burst metadata objects to be included with pattern
        working_dir: Directory to save jsons at

    Returns:
        Directory where jsons were saved
    """
    for key, value in burst_metadata_dict.items():
        with open(working_dir / key, 'w') as json_file:
            json.dump(value, json_file)
    return working_dir


def get_indexes(zipped_safe_path: Path) -> Iterable:
    """Get indexes for XML and bursts in zipped SAFE.

    Args:
        zipped_safe_path: Path to zipped SAFE

    Returns:
        tuple of lists of XmlMetadata and BurstMetadata objects
    """
    with zipfile.ZipFile(zipped_safe_path) as f:
        tiffs = [x for x in f.infolist() if 'tiff' in Path(x.filename).name]
        xmls = [x for x in f.infolist() if 'xml' in Path(x.filename).name]
        xmls += [x for x in f.infolist() if 'manifest.safe' == Path(x.filename).name]

        non_deflate_tiffs = [tiff for tiff in tiffs if tiff.compress_type != zipfile.ZIP_DEFLATED]
        non_deflate_xmls = [xml for xml in xmls if xml.compress_type != zipfile.ZIP_DEFLATED]
        non_deflate = non_deflate_tiffs + non_deflate_xmls
        if non_deflate:
            raise ValueError(
                'Non-deflate compressed files found in SAFE.'
                'All files must be compressed using the deflate method before indexing.'
            )

    print('Reading XMLs...')
    xml_metadatas = [create_xml_metadata(zipped_safe_path, x) for x in tqdm(xmls)]

    print('Reading Bursts...')
    burst_metadatas = []
    for tiff in tqdm(tiffs):
        tiff_name = Path(tiff.filename).name
        annotation_name = Path(tiff_name).with_suffix('.xml').name
        annotation_offset = [item for item in xml_metadatas if item.name == annotation_name][0].offset
        manifest_offset = [item for item in xml_metadatas if item.name == 'manifest.safe'][0].offset
        burst_metadata = create_index(zipped_safe_path, tiff_name, annotation_offset, manifest_offset)
        burst_metadatas.append(burst_metadata)
    burst_metadatas = {k: v for d in burst_metadatas for k, v in d.items()}
    return xml_metadatas, burst_metadatas


def index_safe(slc_name: str, edl_token: str = None, working_dir='.', keep: bool = True):
    """Create the index and other metadata needed to directly download
    and correctly format burst tiffs/metadata Sentinel-1 SAFE zip. Save
    this information in json files. All information for extracting a burst is included in
    the index file.

    Args:
        slc_name: Scene name to index
        edl_token: token for earth data login access
        working_dir: Directory to save metadata and index files
        keep: If False, delete SLC zip after indexing

    Returns:
        No function outputs, but saves a metadata.json, burst indexes to file
    """
    absolute_dir = Path(working_dir).resolve()
    zipped_safe_path = absolute_dir / f'{slc_name}.zip'
    if not zipped_safe_path.exists():
        print('Downloading SLC...')
        utils.download_slc(slc_name, edl_token, working_dir=absolute_dir)
    else:
        print('SLC exists locally, skipping download')

    xml_metadatas, burst_metadatas = get_indexes(zipped_safe_path)

    save_xml_metadata_as_json(xml_metadatas, str(absolute_dir / f'{slc_name}_metadata.json'))
    save_burst_metadata_as_json(burst_metadatas, absolute_dir)

    if not keep:
        os.remove(zipped_safe_path)


def lambda_handler(event, context):
    print('## ENVIRONMENT VARIABLES')
    print(os.environ)
    print('## EVENT')
    print(event)
    print('## PROCESS BEGIN...')
    s3 = boto3.client('s3')
    bucket_name = os.environ.get('IndexBucketName')
    with tempfile.TemporaryDirectory() as tmpdirname:
        index_safe(slc_name=event['scene'], edl_token=event['edl_token'], working_dir=tmpdirname)
        indexes = Path(tmpdirname).glob('*.bstidx')
        [s3.upload_file(str(x), bucket_name, x.name) for x in indexes]
    print('## PROCESS COMPLETE!')


def main():
    """Example Command:

    index_safe.py S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85
    """
    parser = ArgumentParser()
    parser.add_argument('scene')
    args = parser.parse_args()
    index_safe(slc_name=args.scene)


if __name__ == '__main__':
    main()



import json
from pathlib import Path

import lxml.etree as ET
import pytest
from unittest.mock import patch

from index_safe import extract_metadata, utils


def create_minimal_xml(output_path, name):
    root = ET.Element('root')
    name_element = ET.Element('name')
    name_element.text = name
    root.append(name_element)

    tree = ET.ElementTree(root)
    xml_string = ET.tostring(tree, pretty_print=True, encoding='utf-8', xml_declaration=True)
    output_path.write_bytes(xml_string)
    return output_path


def test_extract_metadata_xml_bytes():
    equal_offset = utils.Offset(100, 99)
    negative_offset = utils.Offset(-100, 100)
    with patch('index_safe.utils.http_range_get', return_value=b''):
        client, range_get = utils.setup_download_client(strategy='http')

        with pytest.raises(ValueError, match='offset stop must be greater than offset start'):
            extract_metadata.extract_metadata_xml('https://foo.com', equal_offset, client, range_get)

        with pytest.raises(ValueError, match='offset stop and offset start must be greater than 0'):
            extract_metadata.extract_metadata_xml('https://foo.com', negative_offset, client, range_get)


def test_json_to_metadata_entries(tmpdir):
    test_json = {
        'SLCNAME': {
            'manifest.safe': {'offset_start': 0, 'offset_stop': 100},
            'annotation.xml': {'offset_start': 100, 'offset_stop': 200},
        }
    }
    json_path = Path(tmpdir) / 'test.json'
    json_path.write_text(json.dumps(test_json))
    metadata_entries = extract_metadata.json_to_metadata_entries(json_path)
    assert len(metadata_entries) == 2

    manifest = utils.XmlMetadata('manifest.safe', 'SLCNAME', utils.Offset(0, 100))
    assert metadata_entries[0] == manifest

    annotation = utils.XmlMetadata('annotation.xml', 'SLCNAME', utils.Offset(100, 200))
    assert metadata_entries[1] == annotation


def test_make_input_xml():
    paths = [Path('annotation.xml'), Path('noise.xml'), Path('calibration.xml')]
    manifest = Path('manifest.safe')
    input_xml = extract_metadata.make_input_xml(paths, manifest)
    assert input_xml.tag == 'files'
    assert input_xml.attrib['manifest'] == str(manifest)
    assert len(input_xml) == 3
    for child, path in zip(input_xml, paths):
        assert child.tag == 'file'
        assert child.attrib['name'] == str(path)
        assert child.attrib['label'] in str(paths)


def test_xlst_render(tmpdir):
    tmpdir = Path(tmpdir)
    manifest_path = create_minimal_xml(tmpdir / 'manifest.safe', 'manifest')
    paths = []
    for name in ['annotation', 'calibration', 'noise']:
        paths.append(create_minimal_xml(tmpdir / f'{name}.xml', name))

    renderer = extract_metadata.XsltRenderer(Path(__file__).parent.parent / 'src' / 'index_safe' / 'burst.xsl')
    renderer.render_to(tmpdir / 'burst.xml', paths, manifest_path)

    burst_xml = ET.parse(tmpdir / 'burst.xml')
    child_xmls = burst_xml.findall('.//{*}name')
    assert len(child_xmls) == 4
    assert child_xmls[0].text == 'manifest'
    assert child_xmls[1].text == 'annotation'
    assert child_xmls[2].text == 'calibration'
    assert child_xmls[3].text == 'noise'


def test_select_and_reorder_metadatas():
    unordered_names = [
        's1a-iw2-slc-vh-NULL-NULL-NULL-NULL-NULL.xml',
        's1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        's1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'manifest.safe',
        'noise-s1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'calibration-s1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'noise-s1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'calibration-s1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
    ]

    ordered_names = [
        's1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'noise-s1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'calibration-s1a-iw1-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        's1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'noise-s1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'calibration-s1a-iw2-slc-vv-NULL-NULL-NULL-NULL-NULL.xml',
        'manifest.safe',
    ]

    unordered_metadatas = [utils.XmlMetadata(name, 'SLCNAME', utils.Offset(0, 100)) for name in unordered_names]
    ordered_metadatas = extract_metadata.select_and_reorder_metadatas(unordered_metadatas, 'vv')
    test_names = [metadata.name for metadata in ordered_metadatas]
    assert 's1a-iw2-slc-vh-NULL-NULL-NULL-NULL-NULL.xml' not in test_names
    assert test_names == ordered_names



import base64
import json
import os
import tempfile
import lxml.etree as ET
from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from pathlib import Path
from typing import Callable, Iterable, Tuple, Union

import boto3
import numpy as np
import requests
import zran
from osgeo import gdal

from index_safe.extract_metadata import extract_metadata_xml

try:
    from index_safe import utils
except ModuleNotFoundError:
    import utils


gdal.UseExceptions()

KB = 1024
MB = KB * KB


def extract_burst_data(
    url: str,
    metadata: utils.BurstMetadata,
    index: zran.Index,
    client: Union[boto3.client, requests.sessions.Session],
    range_get_func: Callable,
) -> bytes:
    """Extract bytes pertaining to a burst from a Sentinel-1 SLC archive using a zran index that represents
    a single burst.

    Args:
        url: url location of SLC archive
        metadata: metadata object for burst to extract
        index: zran index object for SLC archive
        client: boto3 S3 client or requests session
        range_get_func: function to use to get a range of bytes from SLC archive

    Returns:
        bytes of burst data
    """
    total_size = (metadata.index_offset.stop - 1) - metadata.index_offset.start
    range_headers = utils.calculate_range_parameters(total_size, metadata.index_offset.start, 20 * MB)
    with ThreadPoolExecutor(max_workers=20) as executor:
        results = executor.map(range_get_func, repeat(client), repeat(url), range_headers)
        body = b''.join(results)
    length = metadata.uncompressed_offset.stop - metadata.uncompressed_offset.start
    burst_bytes = zran.decompress(body, index, metadata.uncompressed_offset.start, length)
    return burst_bytes


def compute_valid_window(index: int, burst: ET.Element) -> utils.Window:
    """Written by Jason Ninneman for the ASF I&A team's burst extractor.
    Using the information contained within a SAFE annotation XML burst
    element, identify the window of a burst that contains valid data.

    Args:
        index: zero-indexed burst number to compute the valid window for
        burst: <burst> element of annotation XML for the given index

    Returns:
        Row and column ranges for the valid data within a burst
    """
    # all offsets, even invalid offsets
    offsets_range = utils.Offset(
        np.array([int(val) for val in burst.find('firstValidSample').text.split()]),
        np.array([int(val) for val in burst.find('lastValidSample').text.split()]),
    )

    # returns the indices of lines containing valid data
    lines_with_valid_data = np.flatnonzero(offsets_range.stop - offsets_range.start)

    # get first and last sample with valid data per line
    # x-axis, range
    valid_offsets_range = utils.Offset(
        offsets_range.start[lines_with_valid_data].min(),
        offsets_range.stop[lines_with_valid_data].max(),
    )

    # get the first and last line with valid data
    # y-axis, azimuth
    valid_offsets_azimuth = utils.Offset(
        lines_with_valid_data.min(),
        lines_with_valid_data.max(),
    )

    # x-length
    length_range = valid_offsets_range.stop - valid_offsets_range.start
    # y-length
    length_azimuth = len(lines_with_valid_data)

    valid_window = utils.Window(
        valid_offsets_range.start,
        valid_offsets_range.start + length_range,
        valid_offsets_azimuth.start,
        valid_offsets_azimuth.start + length_azimuth,
    )

    return valid_window


def get_gcps_from_xml(annotation_xml: ET.Element) -> Iterable[utils.GeoControlPoint]:
    """Get geolocation control points from annotation XML.

    Args:
        annotation_xml: root element of annotation XML

    Returns:
        List of geolocation control points
    """
    xml_gcps = annotation_xml.findall('.//{*}geolocationGridPoint')
    gcps = []
    for xml_gcp in xml_gcps:
        pixel = int(xml_gcp.findtext('.//{*}pixel'))
        line = int(xml_gcp.findtext('.//{*}line'))
        longitude = float(xml_gcp.findtext('.//{*}longitude'))
        latitude = float(xml_gcp.findtext('.//{*}latitude'))
        height = float(xml_gcp.findtext('.//{*}height'))
        gcps.append(utils.GeoControlPoint(pixel, line, longitude, latitude, height))
    return gcps


def format_gcps_for_burst(
    burst_number: int, burst_n_lines: int, swath_gcps: Iterable[utils.GeoControlPoint]
) -> Iterable[utils.GeoControlPoint]:
    """Format geolocation control points for a burst by making line numbers
    relative to the burst, and removing any GCPs that are outside of the burst.

    Args:
        burst_number: zero-indexed burst number
        burst_n_lines: number of lines in the burst
        swath_gcps: list of geolocation control points for the entire swath

    Returns:
        List of geolocation control points for a particular burst
    """
    gcps = []
    burst_starting_line = burst_number * burst_n_lines
    for gcp in swath_gcps:
        gcps.append(utils.GeoControlPoint(gcp.pixel, gcp.line - burst_starting_line, gcp.lon, gcp.lat, gcp.hgt))
    return gcps


def burst_bytes_to_numpy(burst_bytes: bytes, shape: Iterable[int]) -> np.ndarray:
    """Convert bytes representing a burst to numpy array.

    Args:
        burst_bytes: bytes of a burst
        shape: tuple representing shape of the burst array (n_rows, n_cols)

    Returns:
        burst array with a CFloat data type
    """
    array = np.frombuffer(burst_bytes, dtype=np.int16).astype(float)
    array.dtype = 'complex'
    array = array.reshape(shape).astype(np.csingle)
    return array


def invalid_to_nodata(array: np.ndarray, valid_window: utils.Window, nodata_value: int = 0) -> np.ndarray:
    """Use valid window information to set array values outside of valid window to nodata.

    Args:
        array: input burst array to modify
        valid_window: window that will not be set to nodata
        nodata: value used to represent nodata

    Returns
        modified burst array
    """
    is_not_valid = np.ones(array.shape).astype(bool)
    is_not_valid[valid_window.ystart : valid_window.yend, valid_window.xstart : valid_window.xend] = False
    array[is_not_valid] = nodata_value
    return array


def json_to_burst_metadata(burst_json_path: str) -> Tuple[zran.Index, utils.BurstMetadata]:
    """Convert burst metadata json file to a BurstMetadata object and zran index file.

    Args:
        burst_json_path: path to burst metadata json file

    Returns:
        zran index file and BurstMetadata object
    """
    with open(burst_json_path, 'r') as json_file:
        metadata_dict = json.load(json_file)

    shape = (metadata_dict['n_rows'], metadata_dict['n_columns'])
    index_offset = utils.Offset(metadata_dict['index_offset']['start'], metadata_dict['index_offset']['stop'])
    decompressed_offset = utils.Offset(
        metadata_dict['uncompressed_offset']['start'], metadata_dict['uncompressed_offset']['stop']
    )
    annotation_offset = utils.Offset(
        metadata_dict['annotation_offset']['start'], metadata_dict['annotation_offset']['stop']
    )
    manifest_offset = utils.Offset(metadata_dict['manifest_offset']['start'], metadata_dict['manifest_offset']['stop'])
    burst_metadata = utils.BurstMetadata(
        metadata_dict['name'],
        metadata_dict['slc'],
        metadata_dict['swath'],
        metadata_dict['burst_index'],
        shape,
        index_offset,
        decompressed_offset,
        annotation_offset,
        manifest_offset,
    )
    decoded_bytes = base64.b64decode(metadata_dict['dflidx_b64'])
    index = zran.Index.parse_index_file(decoded_bytes)
    return index, burst_metadata


def array_to_raster(
    out_path: Path, array: np.ndarray, gcps: Iterable[utils.GeoControlPoint], fmt: str = 'GTiff'
) -> str:
    """Save a burst array as gdal raster.

    Args:
        out_path: path to save file to
        array: array to save as raster
        gcps: list of gcps to use for georeferencing
        fmt: file format to use

    Returns:
        path to saved raster
    """
    n_rows, n_cols = array.shape
    epsg_4326 = gdal.osr.SpatialReference()
    epsg_4326.ImportFromEPSG(4326)

    driver = gdal.GetDriverByName(fmt)
    out_dataset = driver.Create(str(out_path), n_cols, n_rows, 1, gdal.GDT_CInt16)

    out_dataset.GetRasterBand(1).WriteArray(array)

    out_dataset.SetProjection(epsg_4326.ExportToWkt())

    band = out_dataset.GetRasterBand(1)
    band.SetNoDataValue(0)

    gdal_gcps = [gdal.GCP(point.lon, point.lat, point.hgt, point.pixel, point.line) for point in gcps]
    out_dataset.SetGCPs(gdal_gcps, epsg_4326.ExportToWkt())

    out_dataset = None
    return out_path


def extract_burst(
    burst_index_path: str, edl_token: str = None, strategy: str = 's3', working_dir: Path = Path('.')
) -> str:
    """Extract burst from SLC in ASF archive using a burst-level index
    file. Index must be available locally.

    Args:
        burst_index_path: path to burst index file on disk
        edl_token: EDL token to use for downloading SLC
        strategy: download strategy to use ('s3' | 'https')
        working_dir: directory to use for extract burst to

    Returns:
        path to saved burst raster
    """
    index, burst_metadata = json_to_burst_metadata(burst_index_path)
    url = utils.get_download_url(burst_metadata.slc)

    client, range_get_func = utils.setup_download_client(strategy=strategy)
    burst_bytes = extract_burst_data(url, burst_metadata, index, client, range_get_func)
    annotation_xml = extract_metadata_xml(url, burst_metadata.annotation_offset, client, range_get_func)

    burst_array = burst_bytes_to_numpy(burst_bytes, (burst_metadata.shape))
    valid_window = compute_valid_window(
        burst_metadata.burst_index, annotation_xml.findall('.//{*}burst')[burst_metadata.burst_index]
    )
    burst_array = invalid_to_nodata(burst_array, valid_window)
    swath_gcps = get_gcps_from_xml(annotation_xml)
    gcps = format_gcps_for_burst(burst_metadata.burst_index, burst_metadata.shape[0], swath_gcps)
    out_path = array_to_raster(working_dir / burst_metadata.name, burst_array, gcps)
    return out_path


def lambda_handler(event, context):
    # TODO need to test with new interface
    print('## ENVIRONMENT VARIABLES')
    print(os.environ)
    print('## EVENT')
    print(event)
    print('## PROCESS BEGIN...')
    s3 = boto3.client('s3')
    index_bucket_name = os.environ.get('IndexBucketName')
    extract_bucket_name = os.environ.get('ExtractBucketName')
    burst_json_name = Path(event['burst']).with_suffix('.json')
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = Path(tmpdirname)
        utils.lambda_get_credentials(event['edl_token'], tmpdir, s3, extract_bucket_name, 'credentials.json')
        s3.download_file(index_bucket_name, str(burst_json_name), str(tmpdir / burst_json_name))
        tmp_path = extract_burst(burst_json_name, event['edl_token'], working_dir=tmpdir)
        s3.upload_file(str(tmp_path), extract_bucket_name, tmp_path.name)
    print('## PROCESS COMPLETE!')


def main():
    """Example Command:

    extract_burst.py S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_IW2_VV_0.json
    """
    parser = ArgumentParser()
    parser.add_argument('index_path')
    args = parser.parse_args()

    extract_burst(args.index_path)


if __name__ == '__main__':
    main()



import base64
import json
import re
import zipfile

import zran
import numpy as np
from osgeo import gdal
from pathlib import Path

import pytest
import lxml.etree as ET

from index_safe import utils
from index_safe.create_index import index_safe, save_xml_metadata_as_json
from index_safe.extract_metadata import extract_metadata, json_to_metadata_entries
from index_safe.extract_burst import extract_burst, json_to_burst_metadata


def load_geotiff(infile, band=1):
    ds = gdal.Open(infile, gdal.GA_ReadOnly)

    data = ds.GetRasterBand(band).ReadAsArray()
    nodata = ds.GetRasterBand(band).GetNoDataValue()
    projection = ds.GetProjection()
    transform = ds.GetGeoTransform()
    ds = None
    return data, transform, projection, nodata


@pytest.fixture()
def test_data_dir():
    path = Path(__file__).parent / 'test_data'
    path.mkdir(exist_ok=True)
    return path


@pytest.fixture()
def slc_zip_path(test_data_dir):
    slc_name = 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85'
    slc_path = test_data_dir / f'{slc_name}.zip'
    if slc_path.exists():
        return slc_path

    print('SLC is not present, downloading SLC')
    in_aws = utils.check_if_in_aws_and_region()
    strategy = 's3' if in_aws else 'http'
    utils.download_slc(slc_name, working_dir=test_data_dir, strategy=strategy)
    return slc_path


@pytest.fixture()
def golden_burst_metadata():
    burst_metadata = utils.BurstMetadata(
        name='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_IW2_VV_7.tiff',
        slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
        swath='IW2',
        burst_index=7,
        shape=(1510, 25448),
        index_offset=utils.Offset(start=2987161209, stop=3087580208),
        uncompressed_offset=utils.Offset(start=1140690, stop=154846610),
        annotation_offset=utils.Offset(start=4735212575, stop=4735444360),
        manifest_offset=utils.Offset(start=4735444504, stop=4735449728),
    )
    return burst_metadata


@pytest.fixture()
def golden_burst_index(test_data_dir):
    index_name = 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_IW2_VV_7.dflidx'
    index_path = test_data_dir / index_name
    index = zran.Index.read_file(str(index_path))
    return index


@pytest.fixture()
def golden_burst(test_data_dir, slc_zip_path, golden_burst_metadata):
    swath, polarization = golden_burst_metadata.name.split('_')[-3:-1]
    regex = re.compile(rf'^s1\w-{swath.lower()}-slc-{polarization.lower()}.*tiff$')
    with zipfile.ZipFile(slc_zip_path) as zip_file:
        zinfo = [zinfo for zinfo in zip_file.infolist() if regex.match(Path(zinfo.filename).name)][0]
        if not (test_data_dir / zinfo.filename).exists():
            zip_file.extract(zinfo, path=test_data_dir)

    golden_array = load_geotiff(str(test_data_dir / 'valid_IW2_VV_7.vrt'))[0]
    yield golden_array


@pytest.fixture()
def golden_xml_metadata():
    xml_metadata = [
        utils.XmlMetadata(
            name='manifest.safe',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4735444504, stop=4735449728),
        ),
        utils.XmlMetadata(
            name='s1a-iw1-slc-vv-20200604t022252-20200604t022317-032861-03ce65-004.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4733372170, stop=4733606318),
        ),
        utils.XmlMetadata(
            name='s1a-iw2-slc-vv-20200604t022253-20200604t022318-032861-03ce65-005.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4735212575, stop=4735444360),
        ),
        utils.XmlMetadata(
            name='s1a-iw2-slc-vh-20200604t022253-20200604t022318-032861-03ce65-002.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4732740098, stop=4732971601),
        ),
        utils.XmlMetadata(
            name='noise-s1a-iw1-slc-vv-20200604t022252-20200604t022317-032861-03ce65-004.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4733840607, stop=4733877456),
        ),
        utils.XmlMetadata(
            name='noise-s1a-iw2-slc-vv-20200604t022253-20200604t022318-032861-03ce65-005.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4734619938, stop=4734662287),
        ),
        utils.XmlMetadata(
            name='calibration-s1a-iw1-slc-vv-20200604t022252-20200604t022317-032861-03ce65-004.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4734705983, stop=4734862152),
        ),
        utils.XmlMetadata(
            name='calibration-s1a-iw2-slc-vv-20200604t022253-20200604t022318-032861-03ce65-005.xml',
            slc='S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85',
            offset=utils.Offset(start=4734118441, stop=4734329022),
        ),
    ]
    return xml_metadata


@pytest.mark.integration
def test_create_index(slc_zip_path, golden_xml_metadata, golden_burst_metadata):
    safe_name = slc_zip_path.with_suffix('').name
    test_data_dir = slc_zip_path.parent
    index_safe(safe_name, working_dir=test_data_dir)

    files = list(test_data_dir.glob('S1*.json'))
    metadata = [file for file in files if 'metadata' in file.name]
    indexes = [file for file in files if 'metadata' not in file.name]
    assert len(metadata) == 1
    assert len(indexes) == 54

    metadata_path = metadata[0]
    xml_metadatas = json_to_metadata_entries(metadata_path)
    for golden_xml in golden_xml_metadata:
        test_xml = [xml for xml in xml_metadatas if xml.name == golden_xml.name][0]
        assert test_xml == golden_xml

    index_path = [index for index in indexes if 'IW2_VV_7.json' in index.name][0]
    index, burst_metadata = json_to_burst_metadata(index_path)
    assert golden_burst_metadata == burst_metadata

    golden_index_bytes = index_path.with_suffix('.dflidx').read_bytes()
    assert index.create_index_file() == golden_index_bytes


@pytest.mark.integration
def test_extract_metadata(tmpdir, golden_xml_metadata, slc_zip_path):
    tmpdir = Path(tmpdir)
    json_path = tmpdir / 'metadata.json'
    save_xml_metadata_as_json(golden_xml_metadata, str(json_path))

    in_aws = utils.check_if_in_aws_and_region()
    strategy = 's3' if in_aws else 'http'
    extract_metadata(json_path, 'vv', working_dir=tmpdir, strategy=strategy)
    test_xml = ET.parse(tmpdir / 'transformed.xml')
    manifest, metadata = [child for child in test_xml.getroot()]
    sub_xmls = {child.attrib['source_filename']: child.find('content') for child in metadata}

    with zipfile.ZipFile(slc_zip_path) as zip_file:
        zinfos = zip_file.infolist()
        golden_xmls = []
        for xml_name in sub_xmls:
            zip_path = [zinfo for zinfo in zinfos if xml_name == Path(zinfo.filename).name][0]
            with zip_file.open(zip_path, 'r') as xml_file:
                golden_xmls.append(ET.parse(xml_file).getroot())

    for xml_name, golden_xml in zip(sub_xmls, golden_xmls):
        golden_xml.tag = 'content'
        golden_string = ET.tostring(golden_xml, encoding='unicode')
        golden_string = golden_string.replace(' ', '').replace('\n', '')
        test_string = ET.tostring(sub_xmls[xml_name], encoding='unicode')
        test_string = test_string.replace(' ', '').replace('\n', '')
        assert len(test_string) == len(golden_string)
        assert test_string == golden_string


@pytest.mark.integration
def test_golden_by_burst(tmpdir, golden_burst_metadata, golden_burst_index, golden_burst):
    tmpdir = Path(tmpdir)
    burst_dictionary = golden_burst_metadata.to_dict()
    dflidx = golden_burst_index.create_index_file()
    burst_dictionary['dflidx_b64'] = base64.b64encode(dflidx).decode('utf-8')

    burst_index_path = tmpdir / 'burst_index.json'
    with open(burst_index_path, 'w') as json_file:
        json.dump(burst_dictionary, json_file)

    in_aws = utils.check_if_in_aws_and_region()
    strategy = 's3' if in_aws else 'http'
    burst_path = extract_burst(burst_index_path, strategy=strategy, working_dir=tmpdir)

    test_burst = load_geotiff(str(burst_path))[0]

    equal = np.isclose(golden_burst, test_burst)
    assert np.all(equal)



import base64
import json

import numpy as np
from osgeo import gdal
from pathlib import Path
from unittest.mock import patch

from index_safe import extract_burst, utils


def test_compute_valid_window(annotation_xml):
    golden_window = utils.Window(xstart=10, xend=100, ystart=2, yend=7)
    burst_xmls = annotation_xml.findall('.//{*}burst')
    assert golden_window == extract_burst.compute_valid_window(0, burst_xmls[0])
    assert golden_window == extract_burst.compute_valid_window(1, burst_xmls[1])


def test_get_gcps_from_xml(annotation_xml):
    test_gcps = extract_burst.get_gcps_from_xml(annotation_xml)
    assert len(test_gcps) == 2

    test_gcp = test_gcps[0]
    assert isinstance(test_gcp.pixel, int)
    assert isinstance(test_gcp.lat, float)
    assert test_gcp == utils.GeoControlPoint(0, 0, 50, 20, 1000)


def test_format_gcps_for_burst(annotation_xml):
    burst_index = 1
    burst_length = 10
    test_gcps = extract_burst.get_gcps_from_xml(annotation_xml)
    reformatted_gcps = extract_burst.format_gcps_for_burst(burst_index, burst_length, test_gcps)

    assert reformatted_gcps[0].line == test_gcps[0].line - (burst_index * burst_length)
    assert reformatted_gcps[1].line == test_gcps[1].line - (burst_index * burst_length)


def test_burst_bytes_to_numpy():
    input_bytes = np.array([1, 1, 2, 2, 3, 3, 4, 4], dtype=np.int16).tobytes()
    test_array = extract_burst.burst_bytes_to_numpy(input_bytes, shape=(2, 2))

    golden_array = np.array([[1 + 1j, 2 + 2j], [3 + 3j, 4 + 4j]], dtype=np.csingle)
    assert np.allclose(test_array, golden_array)


def test_invalid_to_nodata():
    input_array = np.ones((4, 4))
    valid_window = utils.Window(xstart=1, xend=3, ystart=1, yend=3)
    nodata_value = 0
    test_array = extract_burst.invalid_to_nodata(input_array, valid_window, nodata_value)
    golden_array = np.array([[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]])
    assert np.allclose(test_array, golden_array)


def test_json_to_burst_metadata(tmpdir):
    json_path = Path(tmpdir) / 'test.json'
    sample_json = {
        'name': 'test_name',
        'slc': 'test_slc',
        'swath': 'test_swath',
        'burst_index': 0,
        'n_rows': 1,
        'n_columns': 2,
        'index_offset': {'start': 3, 'stop': 4},
        'uncompressed_offset': {'start': 5, 'stop': 6},
        'annotation_offset': {'start': 7, 'stop': 8},
        'manifest_offset': {'start': 9, 'stop': 10},
        'dflidx_b64': base64.b64encode(b'0').decode('utf-8'),
    }
    json_path.write_text(json.dumps(sample_json))

    with patch('zran.Index.parse_index_file', return_value=None):
        index, burst_metadata = extract_burst.json_to_burst_metadata(str(json_path))

    golden_metadata = utils.BurstMetadata(
        'test_name',
        'test_slc',
        'test_swath',
        0,
        (1, 2),
        utils.Offset(3, 4),
        utils.Offset(5, 6),
        utils.Offset(7, 8),
        utils.Offset(9, 10),
    )
    assert burst_metadata == golden_metadata


def test_array_to_raster(tmpdir):
    raster_path = Path(tmpdir) / 'test.tiff'
    array = np.ones((3, 3))
    gcps = [utils.GeoControlPoint(0, 1, 2, 3, 4), utils.GeoControlPoint(5, 6, 7, 8, 9)]
    extract_burst.array_to_raster(raster_path, array, gcps)

    ds = gdal.Open(str(raster_path))
    assert ds.RasterXSize == 3
    assert ds.RasterYSize == 3

    epsg_4326 = gdal.osr.SpatialReference()
    epsg_4326.ImportFromEPSG(4326)
    assert ds.GetGCPProjection() == epsg_4326.ExportToWkt()

    gcps = ds.GetGCPs()
    assert len(gcps) == 2
    assert gcps[0].GCPPixel == 0
    assert gcps[1].GCPPixel == 5



import sys
import time

import boto3
import fsspec
import requests


def benchmark(fun, args):
    start = time.time()
    result = fun(**args)
    end = time.time()
    print(f'using {fun.__name__} downloaded {sys.getsizeof(result) * 1e-6:.2f}mb in {end - start:.2f}s')
    return result


def download_requests(url, offset, length):
    range_header = f'bytes={offset}-{offset+length}'

    with requests.session() as client:
        resp = client.get(url, headers={'Range': range_header})
        body = resp.content

    return body


def download_fsspec(url, offset, length):
    base_fs = fsspec.filesystem('https', block_size=5 * (2**20))
    with base_fs.open(url, 'rb') as f:
        f.seek(offset)
        body = f.read(length)
    return body


def download_s3(url, offset, length):
    _, _, bucket, *parts = url.split('/')
    bucket = bucket.split('.')[0]
    key = '/'.join(parts)
    range_header = f'bytes={offset}-{offset+length}'
    client = boto3.client('s3')
    resp = client.get_object(Bucket=bucket, Key=key, Range=range_header)
    body = resp['Body'].read()
    return body


if __name__ == '__main__':
    url_base = 'https://ffwilliams2-shenanigans.s3.us-west-2.amazonaws.com/bursts'
    name = 'S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85.zip'
    url = f'{url_base}/{name}'

    offset = 1_000
    length = 100_000_000

    args_dict = {'url': url, 'offset': offset, 'length': length}

    benchmark(download_requests, args_dict)
    benchmark(download_fsspec, args_dict)
    benchmark(download_s3, args_dict)



import io
import json
import zlib
from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from pathlib import Path
from typing import Callable, Iterable, Optional, Union

import boto3
import lxml.etree as ET
import requests


try:
    from index_safe import utils
except ModuleNotFoundError:
    import utils

KB = 1024
MB = 1024 * KB
MAX_WBITS = 15


def extract_metadata_xml(
    url: str,
    offset: utils.Offset,
    client: Union[boto3.client, requests.sessions.Session],
    range_get_func: Callable,
) -> ET._Element:
    """Extract and decompress bytes pertaining to a metadata XML file from a Sentinel-1 SLC archive.

    Args:
        url: url location of SLC archive
        offset: offset for compressed data range in zip archive
        client: boto3 S3 client or requests session
        range_get_func: function to use to get a range of bytes from SLC archive

    Returns:
        bytes representing metadata xml
    """
    if offset.stop <= offset.start:
        raise ValueError('offset stop must be greater than offset start')
    elif offset.start < 0 or offset.stop < 0:
        raise ValueError('offset stop and offset start must be greater than 0')

    xml_range = f'bytes={offset.start}-{offset.stop-1}'
    compressed_bytes = range_get_func(client, url, xml_range)
    uncompressed_bytes = zlib.decompressobj(-1 * zlib.MAX_WBITS).decompress(compressed_bytes)
    metadata_xml = ET.parse(io.BytesIO(uncompressed_bytes))
    return metadata_xml


def json_to_metadata_entries(json_path: str) -> Iterable[utils.XmlMetadata]:
    """Convert json of XML metadata information to list of XmlMetadata objects.

    Args:
        json_path: path to json file containing xml metadata information

    Returns:
        list of XmlMetadata objects
    """
    with open(json_path, 'r') as json_file:
        metadata_dict = json.load(json_file)

    slc_name = list(metadata_dict.keys())[0]
    metadata_dict = metadata_dict[slc_name]
    xml_metadatas = []
    for key in metadata_dict:
        offset = utils.Offset(metadata_dict[key]['offset_start'], metadata_dict[key]['offset_stop'])
        xml_metadatas.append(utils.XmlMetadata(key, slc_name, offset))
    return xml_metadatas


def make_input_xml(metadata_paths: Iterable[Path], manifest_path: Path) -> ET._Element:
    """Create input xml for xslt transformation.
    Origionally written by Gabe Clark and Rohan Weeden of ASF's Ingest and archive team.

    Args:
        metadata_paths: list of paths to metadata xml files (annotations, calibration, and noise)
        manifest_path: path to manifest.safe file

    Returns:
        lxml.etree._Element representing input xml
    """
    root = ET.Element('files')

    for metadata_path in metadata_paths:
        child = ET.SubElement(root, 'file')
        child.set('name', str(metadata_path))
        child.set('label', metadata_path.name)

    root.attrib['manifest'] = str(manifest_path)

    return root


class XsltRenderer:
    """Class for rendering xml using xslt transformation.
    Origionally written by Jason Ninneman and Rohan Weeden of ASF's Ingest and archive team.
    """

    def __init__(self, template_path: Union[Path, str]):
        self.template_path = Path(template_path)

    def render_to(
        self, out_path: Union[Path, str], metadata_paths: Iterable[Union[Path, str]], manifest_path: Union[Path, str]
    ):
        """Render xml using xslt transformation and save to out_path.

        Args:
            out_path: path to write rendered xml to
            metadata_paths: list of paths to metadata xml files (annotations, calibration, and noise)
            manifest_path: path to manifest.safe file
        """
        out_path = Path(out_path)
        metadata_paths = [Path(p) for p in metadata_paths]
        manifest_path = Path(manifest_path)

        input_xml = make_input_xml(metadata_paths, manifest_path)
        parser = ET.XMLParser(remove_blank_text=True)
        xslt_root = ET.parse(self.template_path, parser)

        transform = ET.XSLT(xslt_root)
        transformed = transform(input_xml)
        transformed.write(out_path, pretty_print=True)


def combine_xml_metadata_files(results_dict: dict, output_name='transformed.xml', directory: Optional[Path] = None):
    """Combine xml metadata files into single xml file using xslt transformation.

    Args:
        results_dict: dictionary where keys are metadata filenames (including manifest.safe)
                      and values are the downloaded xml data as bytes
        output_name: name of combined xml file
        directory: directory to write combined xml file to
    """
    if not directory:
        directory = Path.cwd()
    metadata_paths = []
    for name in results_dict:
        path = directory / name
        if 'manifest' in name:
            manifest_path = path
        else:
            metadata_paths.append(path)
        xml_string = ET.tostring(results_dict[name], pretty_print=True, encoding='utf-8', xml_declaration=True)
        path.write_bytes(xml_string)

    renderer = XsltRenderer(Path(__file__).parent / 'burst.xsl')
    renderer.render_to(directory / output_name, metadata_paths, manifest_path)
    [path.unlink() for path in metadata_paths + [manifest_path]]


def select_and_reorder_metadatas(
    metadatas: Iterable[utils.XmlMetadata], polarization: str
) -> Iterable[utils.XmlMetadata]:
    """Select correct polarization, then reorder XmlMetadata objects swath, then by product, noise, calibration.

    Args:
        metadatas: list of XmlMetadata objects
        polarization: polarization to select (vv | vh | hv | hh)

    Returns:
        list of sorted XmlMetadata objects
    """
    polarization = polarization.lower()
    manifest = [metadata for metadata in metadatas if 'manifest' in metadata.name][0]
    products = []
    noises = []
    calibrations = []
    for metadata in metadatas:
        if metadata.name.startswith('s1') and metadata.name.split('-')[3] == polarization:
            products.append(metadata)
        elif metadata.name.startswith('noise') and metadata.name.split('-')[4] == polarization:
            noises.append(metadata)
        elif metadata.name.startswith('calibration') and metadata.name.split('-')[4] == polarization:
            calibrations.append(metadata)

    products = sorted(products, key=lambda x: x.name.split('-')[1])
    noises = sorted(noises, key=lambda x: x.name.split('-')[2])
    calibrations = sorted(calibrations, key=lambda x: x.name.split('-')[2])

    metadatas = [item for sublist in zip(products, noises, calibrations) for item in sublist]
    metadatas.append(manifest)
    return metadatas


def extract_metadata(json_file_path: str, polarization: str, strategy='s3', working_dir: Optional[Path] = None):
    """Extract all xml metadata files from SLC in ASF archive
    using offset information.

    Args:
        json_file_path: path to json file containing extraction metadata
        polarization: polarization to extract (vv | vh | hv | hh)
        strategy: strategy to use for download (s3 | http) s3 only
            works if runnning from us-west-2 region
    """
    if not working_dir:
        working_dir = Path.cwd()

    metadatas = json_to_metadata_entries(json_file_path)
    metadatas = select_and_reorder_metadatas(metadatas, polarization)
    slc_name = metadatas[0].slc
    url = utils.get_download_url(slc_name)
    offsets = [metadata.offset for metadata in metadatas]

    client, range_get_func = utils.setup_download_client(strategy=strategy)
    with ThreadPoolExecutor(max_workers=20) as executor:
        results = executor.map(extract_metadata_xml, repeat(url), offsets, repeat(client), repeat(range_get_func))

    names = [metadata.name for metadata in metadatas]
    results = {name: result for name, result in zip(names, results)}
    combine_xml_metadata_files(results, directory=working_dir)


def main():
    """Example Command:

    extract_metadata.py S1A_IW_SLC__1SDV_20200604T022251_20200604T022318_032861_03CE65_7C85_metadata.json
    """
    parser = ArgumentParser()
    parser.add_argument('metadata_path')
    parser.add_argument('polarization')
    args = parser.parse_args()

    extract_metadata(args.metadata_path, args.polarization)


if __name__ == '__main__':
    main()



from index_safe import create_index, utils
import zipfile
from unittest.mock import patch


def test_get_burst_annotation_data(annotation_xml):
    burst_shape, burst_offsets = create_index.get_burst_annotation_data(annotation_xml)
    assert burst_shape == (1000, 20000)
    assert len(burst_offsets) == 2
    assert burst_offsets[0] == utils.Offset(10, 110)
    assert burst_offsets[1] == utils.Offset(110, 210)


def test_create_xml_metadata():
    with patch('index_safe.utils.get_zip_compressed_offset', return_value=utils.Offset(0, 100)):
        metadata = create_index.create_xml_metadata('test.zip', zipfile.ZipInfo('test.xml'))

    assert metadata.slc == 'test'
    assert metadata.name == 'test.xml'
    assert metadata.offset == utils.Offset(0, 100)


def test_create_burst_name():
    test_name = 'SLCNAME'
    test_swath = 'foo-iw2-bar-vv-baz.tiff'
    test_burst_index = 1
    burst_name = create_index.create_burst_name(test_name, test_swath, test_burst_index)
    assert burst_name == 'SLCNAME_IW2_VV_1.tiff'


def test_create_burst_dflidx(zip_deflated):
    zip_path, member_name = zip_deflated
    indexer = utils.ZipIndexer(zip_path, member_name, spacing=2**18)
    indexer.create_full_dflidx()

    burst_offset = utils.Offset(5 * 2**20, 10 * 2**20)
    compressed_offset, index_burst_offset, dflidx = create_index.create_burst_dflidx(indexer, burst_offset)
    assert index_burst_offset.start != 0
    assert index_burst_offset.start < burst_offset.start
    assert index_burst_offset.stop < burst_offset.stop

    burst_length = burst_offset.stop - burst_offset.start
    index_burst_length = index_burst_offset.stop - index_burst_offset.start
    assert index_burst_length >= burst_length


def test_create_index(zip_deflated, annotation_xml):
    zip_path, member_name = zip_deflated
    annotation_offset = utils.Offset(0, 100)
    manifest_offset = utils.Offset(100, 200)
    with patch('index_safe.create_index.load_annotation_data', return_value=annotation_xml):
        test_burst_metadata = create_index.create_index(zip_path, member_name, annotation_offset, manifest_offset)

    keys = list(test_burst_metadata.keys())
    first_burst = test_burst_metadata[keys[0]]
    golden_burst = utils.BurstMetadata(
        name='SLCNAME_IW2_VV_0.tiff',
        slc='SLCNAME',
        swath='IW2',
        burst_index=0,
        shape=(1000, 20000),
        index_offset=utils.Offset(0, 0),
        uncompressed_offset=utils.Offset(0, 0),
        annotation_offset=utils.Offset(0, 0),
        manifest_offset=utils.Offset(0, 0),
    )
    assert first_burst['name'] == golden_burst.name
    assert first_burst['slc'] == golden_burst.slc
    assert first_burst['swath'] == golden_burst.swath
    assert first_burst['burst_index'] == golden_burst.burst_index



import json
import os
import boto3

lamb = boto3.client('lambda')

edl_token = os.environ['EDL_TOKEN']
scenes = [
    'S1B_IW_SLC__1SDV_20201115T162313_20201115T162340_024278_02E29D_5C54',
    'S1A_IW_SLC__1SDV_20201203T162353_20201203T162420_035524_042744_6D5C',
]

function_arn = 'ffwilliams2-test-indexer-IndexCreationFunction-qTfV0eMha9PT'

results = []
for scene in scenes:
    payload = {'scene': scene, 'edl_token': edl_token}
    result = lamb.invoke(FunctionName=function_arn, InvocationType='Event', Payload=json.dumps(payload))
    if result['StatusCode'] != 202:
        print(f'Invocation failed with for {scene} with status code {result["StatusCode"]}')
    else:
        print(f'{scene} successfully submitted!')
